import getpass
import logging
import socket
from multiprocessing import current_process

import jpype.dbapi2
from django.db.backends.oracle import base
from sqlalchemy.dialects.oracle.base import OracleDialect

from dj_db_conn_pool.backends.jdbc import JDBCDatabaseWrapperMixin

logger = logging.getLogger(__name__)

oracle_session_info = {
    'v$session.process': str(current_process().pid),
    'v$session.osuser': getpass.getuser(),
    'v$session.machine': socket.gethostname(),
    'v$session.program': 'python',
}


class DatabaseWrapper(JDBCDatabaseWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(OracleDialect):
        def do_ping(self, dbapi_connection):
            try:
                return super().do_ping(dbapi_connection)
            except jpype.dbapi2.DatabaseError:
                return False

    jdbc_driver = 'oracle.jdbc.OracleDriver'

    jdbc_url_prefix = 'jdbc:oracle:thin:@'

    def get_connection_params(self):
        return {
            **oracle_session_info,
            **super().get_connection_params()
        }



import logging

import sqlparams

logger = logging.getLogger(__name__)


class CursorWrapper:
    def __init__(self, cursor):
        self.cursor = cursor

        self._sql_params_converter = sqlparams.SQLParams('pyformat', 'qmark', True, True)

        self.statement = None

    def execute(self, query, params=None):
        if isinstance(params, dict):
            query, params = self._sql_params_converter.format(query, params)
        else:
            query = query.replace('%s', '?')

        self.statement = query

        return self.cursor.execute(query, params)

    def __getattr__(self, attr):
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)



import jpype.dbapi2
from django.db.backends.oracle import base
from sqlalchemy.dialects.oracle.base import OracleDialect

from dj_db_conn_pool.backends.jdbc.oceanbase.mixins import JDBCOceanBaseDatabaseWrapperMixin


class DatabaseWrapper(JDBCOceanBaseDatabaseWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(OracleDialect):
        def do_ping(self, dbapi_connection):
            try:
                return super().do_ping(dbapi_connection)
            except jpype.dbapi2.DatabaseError:
                return False

    def init_connection_state(self):
        # TODO: custom OceanBase (Oracle mode) connection initialization
        pass



import jpype


def patch_all():
    patch_converters()


def patch_converters():
    """
    patch jpype's jdbc converters
    """

    def to_python(value):
        return value._py()

    def to_number(value):
        string = str(value.toString())

        if value.scale() > 0:
            return float(string)

        return int(string)

    @jpype.onJVMStart
    def register_converters():
        from dj_db_conn_pool.backends.jdbc import jdbc_type_converters

        jdbc_type_converters[jpype.java.lang.String] = str
        jdbc_type_converters[jpype.java.sql.Date] = to_python
        jdbc_type_converters[jpype.java.sql.Time] = to_python
        jdbc_type_converters[jpype.java.sql.Timestamp] = to_python
        jdbc_type_converters[jpype.java.math.BigDecimal] = to_number
        jdbc_type_converters[jpype.JArray(jpype.types.JByte)] = bytes
        # jdbc_type_converters[type(None)] = lambda v: v



try:
    # A legacy compatibility wrapper for Unicode handling on Python 2.
    # "ugettext_lazy" has been removed since Django 4.0.
    from django.utils.translation import ugettext_lazy as gettext_lazy
except ImportError:
    # Django 4.0
    from django.utils.translation import gettext_lazy



import logging
import threading

import jpype
import jpype.dbapi2

from dj_db_conn_pool.backends.jdbc.utils import CursorWrapper
from dj_db_conn_pool.core.mixins import PersistentDatabaseWrapperMixin

logger = logging.getLogger(__name__)

jdbc_type_converters = {}

lock_check_jvm_status = threading.Lock()


class JDBCDatabaseWrapperMixin(PersistentDatabaseWrapperMixin):
    @property
    def jdbc_driver(self):
        raise NotImplementedError()

    @property
    def jdbc_url_prefix(self):
        raise NotImplementedError()

    @property
    def jdbc_url(self):
        return '{prefix}//{HOST}:{PORT}/{NAME}'.format(
            prefix=self.jdbc_url_prefix,
            **self.settings_dict
        )

    def create_cursor(self, name=None):
        cursor = self.connection.cursor()
        return CursorWrapper(cursor)

    def get_connection_params(self):
        return self.settings_dict.get('OPTIONS', {})

    def _get_new_connection(self, conn_params):
        with lock_check_jvm_status:
            if not jpype.isJVMStarted():
                jpype.startJVM(ignoreUnrecognized=True)

        conn = jpype.dbapi2.connect(
            self.jdbc_url,
            driver=self.jdbc_driver,
            driver_args=dict(
                user=self.settings_dict['USER'],
                password=self.settings_dict['PASSWORD'],
                **conn_params
            ),
            converters=jdbc_type_converters,
        )

        return conn

    def _close(self):
        if self.connection is not None and self.connection.driver_connection.autocommit:
            # if jdbc connection's autoCommit is on
            # jpype will throw NotSupportedError after rollback called
            # we make a little dynamic patch here, make sure
            # SQLAlchemy will not do rollback before recycling connection
            self.connection._pool._reset_on_return = None

            logger.debug(
                "autoCommit of current JDBC connection to %s %s is on, won't do rollback before returning",
                self.alias, self.connection.driver_connection)

        return super()._close()



from django_tenants.postgresql_backend import base

from dj_db_conn_pool.backends.postgresql.mixins import PGDatabaseWrapperMixin


class DatabaseWrapper(PGDatabaseWrapperMixin, base.DatabaseWrapper):
    """
    Notice: this backend is only available for django_tenants
    """
    pass



import pyodbc
from django.conf import ImproperlyConfigured

from dj_db_conn_pool.core.mixins import PersistentDatabaseWrapperMixin


class ODBCWrapperMixin(PersistentDatabaseWrapperMixin):
    def _get_new_connection(self, conn_params):
        try:
            driver = self.settings_dict['ODBC_OPTIONS']['DRIVER']
        except KeyError:
            raise ImproperlyConfigured('No odbc driver provided')

        conn_str_template = 'DRIVER={DRIVER};SERVER={HOST}:{PORT};DATABASE={NAME};UID={USER};PWD={PASSWORD}'

        connection_string = conn_str_template.format(DRIVER=driver, **self.settings_dict)

        return pyodbc.connect(connection_string)



from django.db.backends.oracle import base
from sqlalchemy.dialects.oracle.base import OracleDialect

from dj_db_conn_pool.backends.odbc.mixins import ODBCWrapperMixin


class DatabaseWrapper(ODBCWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(OracleDialect):
        pass

    def init_connection_state(self):
        pass



import logging

from sqlalchemy import pool

from dj_db_conn_pool.compat import gettext_lazy as _
from dj_db_conn_pool.core import pool_container
from dj_db_conn_pool.core.mixins.creation import DatabaseCreationMixin

logger = logging.getLogger(__name__)


class PersistentDatabaseWrapperMixin:
    def __init__(self, *args, **kwargs):
        # override creation_class
        self.creation_class = type('DatabaseCreationWrapper', (DatabaseCreationMixin, self.creation_class), {})

        super().__init__(*args, **kwargs)

    def __str__(self):
        try:
            conn = repr(self.connection.driver_connection)
        except AttributeError:
            conn = '<Not connected>'

        return f'{self.vendor} connection to {self.alias}: {conn}'

    __repr__ = __str__

    def _set_dbapi_autocommit(self, autocommit):
        self.connection.driver_connection.autocommit = autocommit

    def _set_autocommit(self, autocommit):
        with self.wrap_database_errors:
            try:
                self._set_dbapi_autocommit(autocommit)
            except (Exception,) as exc:
                logger.exception('unable to set autocommit mode of %s(%s) to %s, caused by: %s',
                                 self.vendor, self.alias, autocommit, exc)
                raise exc from None

    def _get_dialect(self):
        return self.SQLAlchemyDialect(dbapi=self.Database)

    def _get_new_connection(self, conn_params):
        return super().get_new_connection(conn_params)

    def get_new_connection(self, conn_params):
        """
        override django.db.backends.<database>.base.DatabaseWrapper.get_new_connection to
        change the default behavior of getting new connection to database, we maintain
        pool_container who contains the connection pool of each database here
        when django call this method to get new connection, we check whether there exists
        the pool of this database(self.alias)
        if the target pool doesn't exist, we will create one
        then grab one connection from the pool and return it to django
        :return:
        """
        with pool_container.lock:
            # acquire the lock, check whether there exists the pool of current database
            # note: the value of self.alias is the name of current database, one of setting.DATABASES
            if not pool_container.has(self.alias):
                # self.alias's pool doesn't exist, time to create it

                # parse parameters of current database from self.settings_dict
                pool_setting = {
                    # transform the keys in POOL_OPTIONS to lower case
                    # to fit sqlalchemy.pool.QueuePool's arguments requirement
                    key.lower(): value
                    # traverse POOL_OPTIONS to get arguments
                    for key, value in
                    # self.settings_dict was created by Django
                    # is the connection parameters of self.alias
                    self.settings_dict.get('POOL_OPTIONS', {}).items()
                    # There are some limits of self.alias's pool's option(POOL_OPTIONS):
                    # the keys in POOL_OPTIONS must be capitalised
                    # and the key's lowercase must be in pool_container.pool_default_params
                    if key == key.upper() and key.lower() in pool_container.pool_default_params
                }

                # replace pool_params's items with pool_setting's items
                # to import custom settings
                pool_params = {
                    **pool_container.pool_default_params,
                    **pool_setting
                }

                # now we have all parameters of self.alias
                # create self.alias's pool
                alias_pool = pool.QueuePool(
                    lambda: self._get_new_connection(conn_params),
                    # SQLAlchemy use the dialect to maintain the pool
                    dialect=self._get_dialect(),
                    # parameters of self.alias
                    **pool_params
                )

                logger.debug(_("%s's pool has been created, parameter: %s"), self.alias, pool_params)

                # pool has been created
                # put into pool_container for reusing
                pool_container.put(self.alias, alias_pool)

        # get self.alias's pool from pool_container
        db_pool = pool_container.get(self.alias)
        # get one connection from the pool
        conn = db_pool.connect()

        logger.debug(
            _("got %s's connection %s from its pool"),
            self.alias, conn.driver_connection)

        return conn

    def close(self, *args, **kwargs):
        logger.debug(
            _("release %s's connection %s to its pool"),
            self.alias, getattr(self.connection, 'driver_connection', None))

        return super().close(*args, **kwargs)



from dj_db_conn_pool.core import pool_container


class DatabaseCreationMixin:
    def _destroy_test_db(self, test_database_name, verbosity):
        # dispose all pools before destroying testdb
        pool_container.dispose()

        # destroy testdb
        return super()._destroy_test_db(test_database_name, verbosity)



import jpype.dbapi2
from django.db.backends.mysql import base
from sqlalchemy.dialects.mysql.base import MySQLDialect

from dj_db_conn_pool.backends.jdbc.oceanbase.mixins import JDBCOceanBaseDatabaseWrapperMixin


class DatabaseWrapper(JDBCOceanBaseDatabaseWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(MySQLDialect):
        def do_ping(self, dbapi_connection):
            try:
                return super().do_ping(dbapi_connection)
            except jpype.dbapi2.DatabaseError:
                return False



import importlib

try:
    importlib.import_module("psycopg")
    from sqlalchemy.dialects.postgresql.psycopg import PGDialect_psycopg as PGDialect
except ModuleNotFoundError:
    from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2 as PGDialect


from dj_db_conn_pool.core.mixins import PersistentDatabaseWrapperMixin


class PGDatabaseWrapperMixin(PersistentDatabaseWrapperMixin):
    class SQLAlchemyDialect(PGDialect):
        pass

    def get_new_connection(self, conn_params):
        connection = super().get_new_connection(conn_params)

        if not connection.info:
            connection.info = connection.driver_connection.info

        return connection



__version__ = '1.2.5'
__author__ = 'Altair Bow'
__author_email__ = 'altair.bow@foxmail.com'
__description__ = 'Persistent database connection backends for Django'


def setup(**kwargs):
    from dj_db_conn_pool.core import pool_container

    for key, value in kwargs.items():
        if key in pool_container.pool_default_params:
            pool_container.pool_default_params[key] = value



from django.contrib.gis.db.backends.postgis import base
from sqlalchemy.dialects.postgresql.base import PGDialect

from dj_db_conn_pool.backends.postgresql.mixins import PGDatabaseWrapperMixin


class DatabaseWrapper(PGDatabaseWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(PGDialect):
        pass



import threading

from dj_db_conn_pool.compat import gettext_lazy as _
from dj_db_conn_pool.core.exceptions import PoolDoesNotExist


class PoolContainer(dict):
    # acquire this lock before modify pool_container
    lock = threading.Lock()

    # the default parameters of pool
    pool_default_params = {
        'pre_ping': True,
        'echo': False,
        'timeout': 30,
        'recycle': 60 * 15,
        'pool_size': 10,
        'max_overflow': 10,
    }

    def has(self, pool_name):
        return pool_name in self

    def put(self, pool_name, pool):
        self[pool_name] = pool

    def get(self, pool_name):
        try:
            return self[pool_name]
        except KeyError:
            raise PoolDoesNotExist(_('No such pool: {pool_name}').format(pool_name=pool_name))

    def dispose(self):
        # dispose all pools
        for name, pool in self.items():
            pool.dispose()


# the pool's container, for maintaining the pools
# every process has it's own pool container
pool_container = PoolContainer()



from dj_db_conn_pool.backends.jdbc import JDBCDatabaseWrapperMixin


class JDBCOceanBaseDatabaseWrapperMixin(JDBCDatabaseWrapperMixin):
    vendor = 'OceanBase'

    jdbc_driver = 'com.alipay.oceanbase.jdbc.Driver'

    jdbc_url_prefix = 'jdbc:oceanbase:'



from django.db.backends.oracle import base
from sqlalchemy.dialects.oracle.base import OracleDialect

from dj_db_conn_pool.core.mixins import PersistentDatabaseWrapperMixin


class DatabaseWrapper(PersistentDatabaseWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(OracleDialect):
        pass

    def _set_dbapi_autocommit(self, autocommit):
        self.connection.driver_connection.autocommit = int(autocommit)



from django.db.backends.postgresql import base

from dj_db_conn_pool.backends.postgresql.mixins import PGDatabaseWrapperMixin


class DatabaseWrapper(PGDatabaseWrapperMixin, base.DatabaseWrapper):
    pass



import logging

from django.db.backends.mysql import base
from sqlalchemy.dialects.mysql.base import MySQLDialect

from dj_db_conn_pool.core.mixins import PersistentDatabaseWrapperMixin

logger = logging.getLogger(__name__)


class DatabaseWrapper(PersistentDatabaseWrapperMixin, base.DatabaseWrapper):
    class SQLAlchemyDialect(MySQLDialect):
        pass

    def _set_dbapi_autocommit(self, autocommit):
        self.connection.driver_connection.autocommit(autocommit)
